# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
# Note: to be removed with MDAnalysis.analysis.hbonds.hbond_analysis in 2.0

from __future__ import print_function, absolute_import


import MDAnalysis
import MDAnalysis.analysis.hbonds
import itertools
import pytest
from MDAnalysis import SelectionError

from numpy.testing import (
    assert_equal, assert_array_equal, assert_almost_equal,
    assert_array_almost_equal, assert_allclose,
)
import numpy as np


from six import StringIO

from MDAnalysisTests.datafiles import PDB_helix, GRO, XTC, waterPSF, waterDCD

# For type guessing:
from MDAnalysis.topology.core import guess_atom_type
from MDAnalysis.core.topologyattrs import Atomtypes


def guess_types(names):
    """GRO doesn't supply types, this returns an Attr"""
    return Atomtypes(np.array([guess_atom_type(name) for name in names], dtype=object))


class TestHydrogenBondAnalysis(object):
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(PDB_helix)

    @staticmethod
    @pytest.fixture(scope='class')
    def values(universe):
        return {
            'num_bb_hbonds': universe.atoms.n_residues - universe.select_atoms('resname PRO').n_residues - 4,
            'donor_resid': np.array([5, 6, 8, 9, 10, 11, 12, 13]),
            'acceptor_resnm': np.array(['ALA', 'ALA', 'ALA', 'ALA', 'ALA', 'PRO', 'ALA', 'ALA'], dtype='U4'),
        }

    kwargs = {
        'selection1': 'protein',
        'selection2': 'protein',
        'detect_hydrogens': "distance",
        'distance': 3.0,
        'angle': 150.0,
    }

    @pytest.fixture(scope='class')
    def h(self, universe):
        kw = self.kwargs.copy()
        # kw.update(kwargs)
        h = MDAnalysis.analysis.hbonds.HydrogenBondAnalysis(universe, **kw)
        if kw['detect_hydrogens'] == 'heuristic':
            with pytest.warns(DeprecationWarning):
                h.run(verbose=False)
        else:
            h.run(verbose=False)
        return h

    def test_helix_backbone(self, values, h):

        assert len(h.timeseries[0]) == values['num_bb_hbonds'], "wrong number of backbone hydrogen bonds"
        assert h.timesteps, [0.0]

    def test_generate_table(self, values, h):

        h.generate_table()
        assert len(h.table) == values['num_bb_hbonds'], "wrong number of backbone hydrogen bonds in table"
        assert isinstance(h.table, np.core.records.recarray)
        assert_array_equal(h.table.donor_resid, values['donor_resid'])
        assert_array_equal(h.table.acceptor_resnm, values['acceptor_resnm'])

    def test_atoms_too_far(self):
        pdb = '''
ATOM      1  N   LEU     1      32.310  13.778  14.372  1.00  0.00      SYST N 0
ATOM      2  OW  SOL     2       3.024   4.456   4.147  1.00  0.00      SYST H 0'''

        u = MDAnalysis.Universe(StringIO(pdb), format="pdb")
        h = MDAnalysis.analysis.hbonds.HydrogenBondAnalysis(u, 'resname SOL', 'protein')
        h.run(verbose=False)
        assert h.timeseries == [[]]

    def test_acceptor_OC1_OC2(self):
        gro = '''test
3
    1ALA    OC1    1   0.000   0.000   0.000
    2ALA      N    2   0.300   0.000   0.000
    2ALA     H1    3   0.200   0.000   0.000
7.29748 7.66094 9.82962'''

        u = MDAnalysis.Universe(StringIO(gro), format="gro")
        h = MDAnalysis.analysis.hbonds.HydrogenBondAnalysis(u)
        h.run(verbose=False)
        assert h.timeseries[0][0][2] == 'ALA2:H1'

    def test_true_traj(self):
        u = MDAnalysis.Universe(GRO, XTC)
        u.add_TopologyAttr(guess_types(u.atoms.names))
        h = MDAnalysis.analysis.hbonds.HydrogenBondAnalysis(u,'protein','resname ASP', distance=3.0, angle=120.0)
        h.run()
        assert len(h.timeseries) == 10

    def test_count_by_time(self, values, h):

        c = h.count_by_time()
        assert c.tolist(), [(0.0, values['num_bb_hbonds'])]

    def test_count_by_type(self, values, h):

        c = h.count_by_type()
        assert_equal(c.frequency, values['num_bb_hbonds'] * [1.0])

    # TODO: Rename
    def test_count_by_type2(self, values, h):

        t = h.timesteps_by_type()
        assert_equal(t.time, values['num_bb_hbonds'] * [0.0])

class TestHydrogenBondAnalysisPBC(TestHydrogenBondAnalysis):
    # This system is identical to above class
    # But has a box introduced, and atoms moved into neighbouring images
    # The results however should remain identical if PBC is used
    # If pbc:True in kwargs is changed, these tests should fail
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        u = MDAnalysis.Universe(PDB_helix)
        # transfer to memory to changes to coordinates are reset
        u.transfer_to_memory()
        # place in huge oversized box
        # real system is about 30A wide at most
        boxsize = 150.
        box = np.array([boxsize, boxsize, boxsize, 90., 90., 90.])
        u.dimensions = box

        # then scatter the atoms throughout various images of this box
        u.atoms[::4].translate([boxsize * 2, 0, 0])
        u.atoms[1::4].translate([0, boxsize *4, 0])
        u.atoms[2::4].translate([-boxsize * 5, 0, -boxsize * 2])

        return u

    kwargs = {
        'selection1': 'protein',
        'selection2': 'protein',
        'detect_hydrogens': "distance",
        'distance': 3.0,
        'angle': 150.0,
        'pbc': True,
    }



class TestHydrogenBondAnalysisHeuristic(TestHydrogenBondAnalysis):
        kwargs = {
            'selection1': 'protein',
            'selection2': 'protein',
            'detect_hydrogens': "heuristic",
            'distance': 3.0,
            'angle': 150.0,
        }


class TestHydrogenBondAnalysisHeavy(TestHydrogenBondAnalysis):

    kwargs = {
        'selection1': 'protein',
        'selection2': 'protein',
        'detect_hydrogens': "heuristic",
        'distance': 3.5,
        'angle': 150.0,
        'distance_type': 'heavy',
    }


class TestHydrogenBondAnalysisHeavyFail(TestHydrogenBondAnalysisHeavy):
    kwargs = {
        'selection1': 'protein',
        'selection2': 'protein',
        'detect_hydrogens': "heuristic",
        'distance': 3.0,
        'angle': 150.0,
        'distance_type': 'heavy',
    }

    @staticmethod
    @pytest.fixture()
    def values(universe):
        return {
            'num_bb_hbonds': 0,  # no H-bonds with a D-A distance < 3.0 A (they start at 3.05 A)
            'donor_resid': np.array([]),
            'acceptor_resnm': np.array([], dtype="<U3"),
        }


class TestHydrogenBondAnalysisChecking(object):
    def _setUp(self):
        self.universe = u = MDAnalysis.Universe(PDB_helix)
        self.kwargs = {
            'selection1': 'protein',
            'selection2': 'protein',
            'detect_hydrogens': "distance",
            'distance': 3.0,
            'angle': 150.0,
        }

    # ignore SelectionWarning
    @pytest.mark.filterwarnings('ignore')
    def _run(self, **kwargs):
        kw = kwargs.copy()
        kw.update(kwargs)
        h = MDAnalysis.analysis.hbonds.HydrogenBondAnalysis(self.universe, **kw)
        h.run(verbose=False)
        return h

    def run_HBA_no_update(self, s1, s2, s1type):
        """test that HydrogenBondAnalysis() raises SelectionError for missing donors/acceptors"""
        # no donors/acceptors; only raises error if no updates
        return self._run(
            selection1=s1,
            selection2=s2,
            update_selection1=False,
            update_selection2=False,
            selection1_type=s1type,
        )

    def runOK(self, s1, s2, s1type):
        """test that HydrogenBondAnalysis() works for protein/protein"""
        try:
            h = self.run_HBA_no_update(s1, s2, s1type)
        except:
            raise pytest.fail("HydrogenBondAnalysis protein/protein failed")
        else:
            return True

    @pytest.mark.parametrize('s1, s2, s1type', itertools.product(
        ("protein", "resname ALA and not backbone"),
        ("protein", "resname ALA and not backbone"),
        ("donor", "acceptor", "both")
    ))
    def test_check_static_selections_pt(self, s1, s2, s1type):
        self._setUp()
        protein = "protein"

        if s1 == s2 == protein:
            self.runOK(s1, s2, s1type)
        else:
            with pytest.raises(SelectionError):
                self.run_HBA_no_update(s1, s2, s1type)

    def run_HBA_update(self, s1, s2, s1type):
        # no donors/acceptors; should not raise error because updates=True
        return self._run(
            selection1=s1,
            selection2=s2,
            update_selection1=True,
            update_selection2=True,
            selection1_type=s1type,
        )

    def run_HBA_dynamic_selections(self, *args):
        try:
            h = self.run_HBA_update(*args)
        except:
            raise pytest.fail("HydrogenBondAnalysis with update=True failed")
        else:
            return True

    @pytest.mark.parametrize('s1, s2, s1type', itertools.product(
        ("protein", "resname ALA and not backbone"),
        ("protein", "resname ALA and not backbone"),
        ("donor", "acceptor", "both")
    ))
    def test_run_empty_selections_pt(self, s1, s2, s1type):
        self._setUp()

        self.run_HBA_dynamic_selections(s1, s2, s1type)


class TestHydrogenBondAnalysisTIP3P(object):
    @staticmethod
    @pytest.fixture()
    def universe():
        return MDAnalysis.Universe(waterPSF, waterDCD)

    kwargs = {
        'selection1': 'all',
        'selection2': 'all',
        'detect_hydrogens': "distance",
        'distance': 3.0,
        'angle': 120.0,
    }

    @pytest.fixture()
    def h(self, universe):
        h = MDAnalysis.analysis.hbonds.HydrogenBondAnalysis(universe, **self.kwargs)
        h.run(verbose=False)
        h.generate_table()
        return h

    @pytest.fixture()
    def normalized_timeseries(self, h):
        # timeseries in normalized form: (t, d_indx1, a_indx1, d_index0, a_index0, donor, acceptor, dist, angle)
        #                   array index:  0     1        2        3         4        5      6        7      8
        timeseries = [[t] + item
                      for t, hframe in zip(h.timesteps, h.timeseries)
                      for item in hframe]
        return timeseries

    # keys are the names in the h.table
    reference = {
        'distance': {'mean': 2.0208776, 'std': 0.31740859},
        'angle': {'mean': 155.13521, 'std': 12.98955},
    }
    
    @pytest.fixture()
    def reference_table(self, normalized_timeseries):
        # reference values for the table only
        return {
            'donor_resnm': ["TIP3"] * len(normalized_timeseries),
            'acceptor_resnm': ["TIP3"] * len(normalized_timeseries),
        }

    # index into timeseries (ADJUST ONCE donor_idx and acceptor_ndx are removed)
    # with keys being field names in h.table
    columns = {
        'time': 0,
        'donor_index': 1,
        'acceptor_index': 2,
        'distance': 5,
        'angle': 6,
    }

    # hackish way to allow looping over self.reference and generating tests
    _functions = {
        'mean': np.mean,
        'std': np.std,
    }

    def test_timeseries(self, h, normalized_timeseries):
        h = h
        assert len(h.timeseries) == 10
        assert len(normalized_timeseries) == 29

        for observable in self.reference:
            idx = self.columns[observable]
            for quantity, reference in self.reference[observable].items():
                func = self._functions[quantity]
                assert_allclose(
                    func([item[idx] for item in normalized_timeseries]), reference,
                    rtol=1e-5, atol=0,
                    err_msg="{quantity}({observable}) does not match reference".format(**vars())
                )

    def test_table_atoms(self, h, normalized_timeseries, reference_table):
        h = h
        table = h.table

        assert len(h.table) == len(normalized_timeseries)

        # test that timeseries and table agree on index data and
        # hydrogen bond information at atom level
        for name, idx in self.columns.items():
            assert_array_almost_equal(table.field(name), [data[idx] for data in normalized_timeseries],
                                      err_msg="table[{name}] and timeseries[{idx} do not agree".format(**vars()))

        # test at residue level (issue #801
        # https://github.com/MDAnalysis/mdanalysis/issues/801)
        for name, ref in reference_table.items():
            assert_array_equal(h.table.field(name), ref, err_msg="resname for {0} do not match (Issue #801)")


class TestHydrogenBondAnalysisAutocorrelation(object):
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(waterPSF, waterDCD)

    @staticmethod
    @pytest.fixture(scope='class')
    def selection():
        return "byres name OH2"

    def test_HydrogenBondLifetimes(self, universe, selection):
        hbonds = MDAnalysis.analysis.hbonds.HydrogenBondAnalysis(universe,
                                                                selection,
                                                                selection,
                                                                distance=3.5,
                                                                angle=120.0)
        hbonds.run(start=0, stop=4)
        hbonds.autocorrelation(tau_max=3)
        assert_almost_equal(hbonds.acf_timeseries[3], 0.75)

    def test_HydrogenBondLifetimes_growing_continuous(self, universe, selection):
        hbonds = MDAnalysis.analysis.hbonds.HydrogenBondAnalysis(universe,
                                                                selection,
                                                                selection,
                                                                distance=3.5,
                                                                angle=120.0)
        hbonds.run(start=0, stop=9)
        hbonds.autocorrelation(tau_max=5)
        assert all([frame >= framePlus1 for frame, framePlus1 in zip(hbonds.acf_timeseries, hbonds.acf_timeseries[1:])])



class TestHydrogenBondAnalysisTIP3PHeavyPBC(object):
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(waterPSF, waterDCD)

    kwargs = {
        'selection1': 'all',
        'selection2': 'all',
        'detect_hydrogens': "distance",
        'distance': 3.0,
        'angle': 120.0,
        'distance_type': 'heavy',
        'pbc': True
    }

    @pytest.fixture(scope='class')
    def h(self, universe):
        h = MDAnalysis.analysis.hbonds.HydrogenBondAnalysis(universe, **self.kwargs)
        h.run(verbose=False)
        h.generate_table()
        return h

    @pytest.fixture(scope='class')
    def normalized_timeseries(self, h):
        # timeseries in normalized form: (t, d_indx1, a_indx1, d_index0, a_index0, donor, acceptor, dist, angle)
        #                   array index:  0     1        2        3         4        5      6        7      8
        timeseries = [[t] + item
                      for t, hframe in zip(h.timesteps, h.timeseries)
                      for item in hframe]
        return timeseries

    def test_timeseries(self, h, normalized_timeseries):

        assert len(h.timeseries) == 10
        assert len(normalized_timeseries) == 32

        reference = {
            'distance': {'mean': 2.7627309, 'std': 0.0905052},
            'angle': {'mean': 158.9038039, 'std': 12.0362826},
        }

        assert_allclose(np.mean([hbond[5] for hbond in normalized_timeseries]), reference['distance']['mean'])
        assert_allclose(np.std([hbond[5] for hbond in normalized_timeseries]), reference['distance']['std'])
        assert_allclose(np.mean([hbond[6] for hbond in normalized_timeseries]), reference['angle']['mean'])
        assert_allclose(np.std([hbond[6] for hbond in normalized_timeseries]), reference['angle']['std'])

    def test_count_by_time(self, h):

        ref_times = np.arange(0.02, 0.21, 0.02)
        ref_counts = np.array([3, 2, 4, 4, 4, 4, 3, 2, 3, 3])

        times, counts = np.array([[count[0], count[1]] for count in h.count_by_time()]).T

        assert_array_almost_equal(times, ref_times)
        assert_array_equal(counts, ref_counts)

    def test_count_by_type(self, h):

        ref_values = [1.0, 1.0, 0.5, 0.4, 0.2, 0.1]

        c = h.count_by_type()
        assert_equal(np.sort(c.frequency)[::-1], ref_values)


class TestHydrogenBondAnalysisTIP3PHeavyPBCStartStep(object):
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(waterPSF, waterDCD)

    kwargs = {
        'selection1': 'all',
        'selection2': 'all',
        'detect_hydrogens': "distance",
        'distance': 3.0,
        'angle': 120.0,
        'distance_type': 'heavy',
        'pbc': True
    }

    @pytest.fixture(scope='class')
    def h(self, universe):
        h = MDAnalysis.analysis.hbonds.HydrogenBondAnalysis(universe, **self.kwargs)
        h.run(start=1, step=2, verbose=False)
        h.generate_table()
        return h

    @pytest.fixture(scope='class')
    def normalized_timeseries(self, h):
        # timeseries in normalized form: (t, d_indx1, a_indx1, d_index0, a_index0, donor, acceptor, dist, angle)
        #                   array index:  0     1        2        3         4        5      6        7      8
        timeseries = [[t] + item
                      for t, hframe in zip(h.timesteps, h.timeseries)
                      for item in hframe]
        return timeseries

    def test_timeseries(self, h, normalized_timeseries):

        assert len(h.timeseries) == 5
        assert len(normalized_timeseries) == 15

        reference = {
            'distance': {'mean': 2.73942464, 'std': 0.05867924},
            'angle': {'mean': 157.07768079, 'std': 9.72636682},
        }

        assert_allclose(np.mean([hbond[5] for hbond in normalized_timeseries]), reference['distance']['mean'])
        assert_allclose(np.std([hbond[5] for hbond in normalized_timeseries]), reference['distance']['std'])
        assert_allclose(np.mean([hbond[6] for hbond in normalized_timeseries]), reference['angle']['mean'])
        assert_allclose(np.std([hbond[6] for hbond in normalized_timeseries]), reference['angle']['std'])

    def test_count_by_time(self, h):

        ref_times = np.array([0.04, 0.08, 0.12, 0.16, 0.20])
        ref_counts = np.array([2, 4, 4, 2, 3])

        times, counts = np.array([[count[0], count[1]] for count in h.count_by_time()]).T

        assert_array_almost_equal(times, ref_times)
        assert_array_equal(counts, ref_counts)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from __future__ import absolute_import

import MDAnalysis as mda
import pytest

from MDAnalysisTests.topology.base import ParserBase
from MDAnalysisTests.datafiles import TXYZ, ARC


class TestTXYZParser(ParserBase):
    parser = mda.topology.TXYZParser.TXYZParser
    guessed_attrs = ['masses']
    expected_attrs = ['ids', 'names', 'bonds', 'types']
    expected_n_residues = 1
    expected_n_atoms = 9
    expected_n_segments = 1

    def test_number_of_bonds(self, top):
        assert len(top.bonds.values) == 8

    @pytest.fixture(params=(TXYZ, ARC))
    def filename(self, request):
        return request.param

    def test_atom_type_type(self, top):
        """
        ``u.atoms.types`` contains strings.

        See `issue #2435 <https://github.com/MDAnalysis/mdanalysis/issues/2435>
        """
        types = top.types.values
        type_is_str = [isinstance(atom_type, str) for atom_type in types]
        assert all(type_is_str)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""Tests for the API definition of Timestep


_TestTimestepInterface tests the Readers are correctly using Timesteps
"""
from __future__ import absolute_import

import numpy as np
from numpy.testing import assert_equal

import MDAnalysis as mda
from MDAnalysisTests.datafiles import (PSF, XYZ_five, INPCRD, DCD, DLP_CONFIG,
                                       DLP_HISTORY, DMS, GMS_ASYMOPT, GRO, XTC,
                                       TRR, LAMMPSdata, LAMMPSdata2,
                                       LAMMPSdcd2, mol2_molecules, PDB_small,
                                       PDBQT_input, PQR, PRM, TRJ, PRMncdf,
                                       NCDF, TRZ_psf, TRZ)

from MDAnalysisTests.coordinates.base import BaseTimestepTest, assert_timestep_equal
import pytest

# Can add in custom tests for a given Timestep here!


class TestBaseTimestep(BaseTimestepTest):
    @pytest.mark.parametrize('otherTS', [
        mda.coordinates.TRJ.Timestep,
        mda.coordinates.DMS.Timestep,
        mda.coordinates.GRO.Timestep,
        mda.coordinates.TRZ.Timestep,
    ])
    def test_other_timestep(self, otherTS):
        # use a subclass to base.Timestep to check it works
        ts1 = mda.coordinates.base.Timestep(10)
        ts1.positions = self._get_pos()
        ts2 = otherTS(10)
        ts2.positions = self._get_pos()
        assert_timestep_equal(ts1, ts2, "Failed on {0}".format(otherTS))


# TODO: Merge this into generic Reader tests
# These tests are all included in BaseReaderTest
# Once Readers use that TestClass, delete this one

class TestBaseTimestepInterface(object):
    """Test the Timesteps created by Readers

    This checks that Readers are creating correct Timestep objects,
    ensuring a consistent interface when using Timestep objects.

    Failures here are the Reader's fault

    See Issue #250 for discussion
    """
    @pytest.fixture(params=(
        (XYZ_five, INPCRD, None, None),
        (PSF, DCD, None, None),
        (DLP_CONFIG, None, 'CONFIG', None),
        (DLP_HISTORY, None, 'HISTORY', None),
        (DMS, None, None, None),
        (GRO, None, None, None),
        (XYZ_five, INPCRD, None, None),
        (LAMMPSdata, None, None, None),
        (mol2_molecules, None, None, None),
        (PDB_small, None, None, None),
        (PQR, None, None, None),
        (PDBQT_input, None, None, None),
        (PRM, TRJ, None, None),
        (GRO, XTC, None, None),
        (TRZ_psf, TRZ, None, None),
        (GRO, TRR, None, None),
        (GMS_ASYMOPT, GMS_ASYMOPT, 'GMS', 'GMS'),
        (LAMMPSdata2, LAMMPSdcd2, 'LAMMPS', 'DATA'),
        (PRMncdf, NCDF, None, None),
    ))
    def universe(self, request):
        topology, trajectory, trajectory_format, topology_format = request.param
        if trajectory_format is not None and topology_format is not None:
            return mda.Universe(topology, trajectory, format=trajectory_format,
                                topology_format=topology_format)

        if trajectory is not None:
            return mda.Universe(topology, trajectory)
        else:
            return mda.Universe(topology, format=trajectory_format)

    def test_frame(self, universe):
        assert_equal(universe.trajectory.ts.frame, 0)

    def test_dt(self, universe):
        assert_equal(universe.trajectory.dt, universe.trajectory.ts.dt)


@pytest.mark.parametrize('uni', [
    [(PSF, DCD), {}],  # base Timestep
    [(DLP_CONFIG,), {'format': 'CONFIG'}],  # DLPoly
    [(DMS,), {}],  # DMS
    [(GRO,), {}],  # GRO
    [(np.zeros((1, 30, 3)),), {}],  # memory
    [(PRM, TRJ), {}],  # TRJ
    [(TRZ_psf, TRZ), {}],  # TRZ
])
def test_atomgroup_dims_access(uni):
    uni_args, uni_kwargs = uni
    # check that AtomGroup.dimensions always returns a copy
    u = mda.Universe(*uni_args, **uni_kwargs)

    ag = u.atoms[:10]

    dims = ag.dimensions
    ts = u.trajectory.ts

    # dimensions from AtomGroup should be equal
    assert_equal(dims, ts.dimensions)
    # but not identical
    assert dims is not ts.dimensions



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from __future__ import absolute_import
import re

import numpy as np
import pytest
from numpy.testing import assert_equal, assert_almost_equal

import MDAnalysis as mda
import MDAnalysis.analysis.helanal
from MDAnalysisTests.datafiles import (GRO, XTC, PSF, DCD, PDB_small,
                                       HELANAL_BENDING_MATRIX,
                                       HELANAL_BENDING_MATRIX_SUBSET)

# reference data from a single PDB file:
#   data = MDAnalysis.analysis.helanal.helanal_main(PDB_small,
#                    select="name CA and resnum 161-187")
HELANAL_SINGLE_DATA = {
    'Best fit tilt': 1.3309656332019535,
    'Height': np.array([ 1.5286051 ,  0.19648294,  0.11384312],
                       dtype=np.float32),
    'Local bending angles':
        np.array([  3.44526005,   4.85425806,   4.69548464,   2.39473653,
                    3.56172442,   3.97128344,   3.41563916,   1.86140978,
                    5.22997046,   5.41381264,  27.49601364,  39.69839478,
                    35.05921936,  21.78928566,   9.8632431 ,   8.80066967,
                    5.5344553 ,   6.14356709,  10.15450764,  11.07686138,
                    9.23541832], dtype=np.float32),
    'Residues/turn': np.array([ 3.64864163,  0.152694  ,  0.1131402 ]),
    'Rotation Angles':
        np.array([  87.80540079, -171.86019984,  -75.2341296 ,   24.61695962,
                    121.77104796, -134.94786976,  -35.07857424,   58.9621866 ,
                    159.40210233, -104.83368122,   -7.54816243,   87.40202629,
                    -176.13071955,  -89.13196878,   17.17321345,  105.26627814,
                    -147.00075298,  -49.36850775,   54.24557615,  156.06486532,
                    -110.82698327,   -5.72138626,   85.36050546, -167.28218858,
                    -68.23076936]),
    'Twist': np.array([ 98.83011627,   4.08171701,   3.07253003],
                      dtype=np.float32),
    'Unit twist angles':
        np.array([  97.23709869,   99.09676361,   97.25350952,  101.76019287,
                    100.42689514,   97.08784485,   97.07430267,   98.33553314,
                    97.86578369,   95.45792389,   97.10089111,   95.26415253,
                    87.93136597,  108.38458252,   95.27167511,  104.01931763,
                    100.7199707 ,  101.48034668,   99.64170074,   94.78946686,
                    102.50147247,   97.25154877,  104.54204559,  101.42829895],
                 dtype=np.float32),
    }


def read_bending_matrix(fn):
    """Read helanal_bending_matrix.dat into dict of numpy arrays.

    This is a quick and dirty parser with no error checking.

    Format::

      Mean
         0.0    5.7   10.9 ....
         .

      SD
        ....
        .

      ABDEV
       ....
       .

    Returns
    -------
       data : dict
           dictionary with keys 'Mean', 'SD', 'ABDEV' and NxN matrices.
    """
    data = {}
    current = None
    with open(fn) as bendingmatrix:
        for line in bendingmatrix:
            line = line.strip()
            if not line:
                continue
            if re.match("\D+", line):
                label = line.split()[0]
                current = data[label] = []
            else:
                current.append([float(x) for x in line.split()])
    for k,v in data.items():
        data[k] = np.array(v)
    return data


def test_helanal_trajectory(tmpdir, reference=HELANAL_BENDING_MATRIX,
                            outfile="helanal_bending_matrix.dat"):
    u = mda.Universe(PSF, DCD)
    with tmpdir.as_cwd():
        # Helix 8: 161 - 187 http://www.rcsb.org/pdb/explore.do?structureId=4AKE
        MDAnalysis.analysis.helanal.helanal_trajectory(
            u, select="name CA and resnum 161-187")
        bendingmatrix = read_bending_matrix(outfile)
        ref = read_bending_matrix(reference)
        assert_equal(sorted(bendingmatrix.keys()), sorted(ref.keys()),
                     err_msg="different contents in bending matrix data file")
        for label in ref.keys():
            assert_almost_equal(
                bendingmatrix[label], ref[label], err_msg="bending matrix "
                "stats for {0} mismatch".format(label))


def test_helanal_trajectory_slice(tmpdir,
                                  reference=HELANAL_BENDING_MATRIX_SUBSET,
                                  outfile="helanal_bending_matrix.dat"):
    """Simple regression test to validate that begin/finish work as
    intended. In this case, frames 10 (time: 10.999999031120204) through to
    79 (time: 79.99999295360149) should be picked up."""
    u = mda.Universe(PSF, DCD)
    with tmpdir.as_cwd():
        # Helix 8: 161 - 187 http://www.rcsb.org/pdb/explore.do?structureId=4AKE
        MDAnalysis.analysis.helanal.helanal_trajectory(
            u, select="name CA and resnum 161-187", begin=10, finish=80
        )
        bendingmatrix = read_bending_matrix(outfile)
        ref = read_bending_matrix(reference)
        assert_equal(sorted(bendingmatrix.keys()), sorted(ref.keys()),
                     err_msg="different contents in bending matrix data file")
        for label in ref.keys():
            assert_almost_equal(
                bendingmatrix[label], ref[label], err_msg="bending matrix "
                "starts for {0} mismatch".format(label))


def test_helanal_main(reference=HELANAL_SINGLE_DATA):
    u = mda.Universe(PDB_small)
    # Helix 8: 161 - 187 http://www.rcsb.org/pdb/explore.do?structureId=4AKE
    data = MDAnalysis.analysis.helanal.helanal_main(
        PDB_small, select="name CA and resnum 161-187")
    ref = reference
    assert_equal(sorted(data.keys()), sorted(ref.keys()),
                     err_msg="different contents in data dict")
    for label in ref.keys():
        assert_almost_equal(data[label], ref[label], decimal=4,
                                  err_msg="data[{0}] mismatch".format(label))


def test_exceptions(tmpdir):
    """Testing exceptions which can be raised"""
    u = MDAnalysis.Universe(GRO, XTC)
    u.trajectory[1]

    # Testing xtc striding: Check for resolution of Issue #188
    with tmpdir.as_cwd():
        with pytest.raises(ValueError):
            MDAnalysis.analysis.helanal.helanal_trajectory(
                u, select="name CA", finish=5
            )

    with tmpdir.as_cwd():
        with pytest.raises(ValueError):
            MDAnalysis.analysis.helanal.helanal_trajectory(
                u, select="name CA", begin=1, finish=0
            )

    with tmpdir.as_cwd():
        with pytest.raises(ValueError):
            MDAnalysis.analysis.helanal.helanal_trajectory(
                u, select="name CA", begin=99999
            )


def test_warnings(tmpdir):
    """Testing that a warning which can be raised"""
    u = MDAnalysis.Universe(GRO, XTC)
    u.trajectory[1]

    with tmpdir.as_cwd():
        with pytest.warns(UserWarning) as record:
            MDAnalysis.analysis.helanal.helanal_trajectory(
                u, select="name CA", begin=-1, finish=99999
            )

            assert len(record) == 3
            wmsg1 = ("The input begin time (-1 ps) precedes the starting "
                   "trajectory time --- Setting starting frame to 0".format(
                    -1,0))
            assert str(record[1].message.args[0]) == wmsg1
            wmsg2 = ("The input finish time ({0} ps) occurs after the end of "
                     "the trajectory ({1} ps). Finish time will be set to "
                     "the end of the trajectory".format(
                      99999,1000.0000762939453))
            assert str(record[2].message.args[0]) == wmsg2



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from __future__ import absolute_import

import pytest
from six.moves import cPickle

import MDAnalysis as mda
from numpy.testing import (
    TestCase,
    assert_equal
)

import gc

from MDAnalysisTests.datafiles import PDB_small


class TestAtomGroupPickle(object):
    """Set up hopefully unique universes."""

    # _n marks named universes/atomgroups/pickled strings

    @staticmethod
    @pytest.fixture()
    def universe():
        return mda.Universe(PDB_small, PDB_small, PDB_small)

    @staticmethod
    @pytest.fixture()
    def universe_n():
        return mda.Universe(PDB_small, PDB_small, PDB_small, anchor_name="test1")

    @staticmethod
    @pytest.fixture()
    def ag(universe):
        return universe.atoms[:20]

    @staticmethod
    @pytest.fixture()
    def ag_n(universe_n):
        return universe_n.atoms[:10]

    @staticmethod
    @pytest.fixture()
    def pickle_str(ag):
        return cPickle.dumps(ag, protocol=cPickle.HIGHEST_PROTOCOL)

    @staticmethod
    @pytest.fixture()
    def pickle_str_n(ag_n):
        return cPickle.dumps(ag_n, protocol=cPickle.HIGHEST_PROTOCOL)

    def test_unpickle(self, pickle_str, ag, universe):
        """Test that an AtomGroup can be unpickled (Issue 293)"""
        newag = cPickle.loads(pickle_str)
        # Can unpickle
        assert_equal(ag.indices, newag.indices)
        assert newag.universe is universe, "Unpickled AtomGroup on wrong Universe."

    def test_unpickle_named(self, pickle_str_n, ag_n, universe_n):
        """Test that an AtomGroup can be unpickled (Issue 293)"""
        newag = cPickle.loads(pickle_str_n)
        # Can unpickle
        assert_equal(ag_n.indices, newag.indices)
        assert newag.universe is universe_n, "Unpickled AtomGroup on wrong Universe."

    def test_unpickle_missing(self):
        universe = mda.Universe(PDB_small, PDB_small, PDB_small)
        universe_n = mda.Universe(PDB_small, PDB_small, PDB_small,
                                  anchor_name="test1")
        ag = universe.atoms[:20]  # prototypical AtomGroup
        ag_n = universe_n.atoms[:10]
        pickle_str_n = cPickle.dumps(ag_n, protocol=cPickle.HIGHEST_PROTOCOL)
        # Kill AtomGroup and Universe
        del ag_n
        del universe_n
        # and make sure they're very dead
        gc.collect()
        # we shouldn't be able to unpickle
        # assert_raises(RuntimeError, cPickle.loads, pickle_str_n)
        with pytest.raises(RuntimeError):
            cPickle.loads(pickle_str_n)

    def test_unpickle_noanchor(self, universe, pickle_str):
        # Shouldn't unpickle if the universe is removed from the anchors
        universe.remove_anchor()
        # In the complex (parallel) testing environment there's the risk of
        # other compatible Universes being available for anchoring even after
        # this one is expressly removed.
        # assert_raises(RuntimeError, cPickle.loads, pickle_str)
        with pytest.raises(RuntimeError):
            cPickle.loads(pickle_str)
        # If this fails to raise an exception either:
        # 1-the anchoring Universe failed to remove_anchor or 2-another
        # Universe with the same characteristics was created for testing and is
        # being used as anchor."

    def test_unpickle_reanchor(self, universe, pickle_str, ag):
        # universe is removed from the anchors
        universe.remove_anchor()
        # now it goes back into the anchor list again
        universe.make_anchor()
        newag = cPickle.loads(pickle_str)
        assert_equal(ag.indices, newag.indices)
        assert newag.universe is universe, "Unpickled AtomGroup on wrong Universe."

    def test_unpickle_wrongname(self, universe_n, pickle_str_n):
        # we change the universe's anchor_name
        universe_n.anchor_name = "test2"
        # shouldn't unpickle if no name matches, even if there's a compatible
        # universe in the unnamed anchor list.
        with pytest.raises(RuntimeError):
            cPickle.loads(pickle_str_n)

    def test_unpickle_rename(self, universe_n, universe, pickle_str_n, ag_n):
        # we change universe_n's anchor_name
        universe_n.anchor_name = "test2"
        # and make universe a named anchor
        universe.anchor_name = "test1"
        newag = cPickle.loads(pickle_str_n)
        assert_equal(ag_n.indices, newag.indices)
        assert newag.universe is universe, "Unpickled AtomGroup on wrong Universe."

    def test_pickle_unpickle_empty(self, universe):
        """Test that an empty AtomGroup can be pickled/unpickled (Issue 293)"""
        ag = universe.atoms[[]]
        pickle_str = cPickle.dumps(ag, protocol=cPickle.HIGHEST_PROTOCOL)
        newag = cPickle.loads(pickle_str)
        assert len(newag) == 0


class TestPicklingUpdatingAtomGroups(object):

    @staticmethod
    @pytest.fixture()
    def u():
        return mda.Universe(PDB_small)

    def test_pickling_uag(self, u):
        ag = u.atoms[:100]
        uag = ag.select_atoms('name C', updating=True)
        pickle_str = cPickle.dumps(uag, protocol=cPickle.HIGHEST_PROTOCOL)
        new_uag = cPickle.loads(pickle_str)

        assert_equal(uag.indices, new_uag.indices)

    def test_pickling_uag_of_uag(self, u):
        uag1 = u.select_atoms('name C or name H', updating=True)
        uag2 = uag1.select_atoms('name C', updating=True)
        pickle_str = cPickle.dumps(uag2, protocol=cPickle.HIGHEST_PROTOCOL)
        new_uag2 = cPickle.loads(pickle_str)

        assert_equal(uag2.indices, new_uag2.indices)
